import pandas as pd
from pandas.tseries.frequencies import infer_freq, to_offset

from .utils import timestamp_converter


class PandasMixin:
    def to_dataframe(self):
        df = pd.DataFrame(data={self.name: self.values()}, index=self.index)

        # Need at least 3 dates to infer frequency
        if len(df.index) >= 3:
            # Try to infer freq if is evenly-spaced to get the df.index.freq
            df.index.freq = infer_freq(df.index)

        return df

    def sample(self,
               freq=None,
               start=None,
               end=None,
               index=None,
               interpolate=None):
        """Sample your timeseries into Evenly Spaced TimeSeries.

        Args:
            freq (timedelta): frequency to convert in.
            start (datetime): left bound. Default to None, which result into
                :meth:`~timeseries.TimeSeries.lower_bound`.
            end (datetime): right bound. Default to None, which result into
                :meth:`~timeseries.TimeSeries.upper_bound`.

        Returns:
            evenly-spaced timeseries.
        """

        if freq is None and index is None:
            msg = ("You should either select one frequency OR an index for "
                   "the sampling. Both are None.")
            raise ValueError(msg)

        if not interpolate:
            interpolate = self._default_interpolate

        freq = to_offset(freq)

        ts = self.__class__(default=self.default)
        if self.empty:
            return ts

        if index is not None:
            for idx in index:
                ts[idx] = self[idx]
            return ts

        if start:
            start = timestamp_converter(start)
            if not self._has_default:
                start = max(start, self.lower_bound)
        else:
            start = self.lower_bound

        if end:
            end = timestamp_converter(end)
        else:
            end = self.upper_bound + freq

        dt = start
        while dt < end:
            ts[dt] = self[dt, interpolate]
            dt = dt + freq

        return ts



import pytest

from ticts import TimeSeries, testing

from .conftest import CURRENT, HALFHOUR, ONEHOUR


class TestTictsAdd:
    def test_simple_add(self, smallts, smalldict):
        ts = smallts + smallts
        newdct = {key: value + value for key, value in smalldict.items()}
        expected_ts = TimeSeries(newdct)
        testing.assert_ts_equal(ts, expected_ts)

    def test_simple_add_one_float(self, smallts):
        ts = smallts + 1000
        assert list(ts.values()) == list(range(1000, 1010))

    def test_add_with_keys_differences_with_default(self, smallts_withdefault,
                                                    otherts_withdefault):
        ts = smallts_withdefault + otherts_withdefault
        assert ts.default == smallts_withdefault.default + otherts_withdefault.default
        assert ts[CURRENT + 1 * ONEHOUR] == 1 + otherts_withdefault.default
        assert ts[CURRENT + 2 * ONEHOUR] == 2 + 1000
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2 + 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 3 + 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 4 + 3000

    def test_add_with_keys_differences_without_default(self, smallts, otherts):
        ts = smallts + otherts
        assert not ts._has_default
        assert CURRENT + 1 * ONEHOUR not in ts.index
        assert ts[CURRENT + 2 * ONEHOUR] == 2 + 1000
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2 + 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 3 + 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 4 + 3000

    def test_add_with_keys_differences_with_mixed_default_nodefault(
            self, smallts_withdefault, otherts):
        ts = smallts_withdefault + otherts
        assert not ts._has_default
        assert CURRENT + 1 * ONEHOUR not in ts.index
        assert ts[CURRENT + 2 * ONEHOUR] == 2 + 1000
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2 + 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 3 + 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 4 + 3000

    def test_add_on_list_of_timeseries(self, smallts):
        lst_timeseries = [smallts, smallts, smallts]
        result = sum(lst_timeseries)

        for key in result:
            assert result[key] == 3 * smallts[key]


class TestTictsSub:
    def test_simple_sub(self, smallts):
        ts = smallts - smallts
        assert all([val == 0 for val in ts.values()])

    def test_simple_sub_one_float(self, smallts):
        ts = smallts - 1
        assert list(ts.values()) == list(range(-1, 9))

    def test_sub_with_keys_differences_with_default(self, smallts_withdefault,
                                                    otherts_withdefault):
        ts = smallts_withdefault - otherts_withdefault
        assert ts[CURRENT + 1 * ONEHOUR] == 1 - 900
        assert ts[CURRENT + 2 * ONEHOUR] == 2 - 1000
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2 - 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 3 - 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 4 - 3000

    def test_sub_with_keys_differences_without_default(self, smallts, otherts):
        ts = smallts - otherts
        assert CURRENT + 1 * ONEHOUR not in ts.index
        assert ts[CURRENT + 2 * ONEHOUR] == 2 - 1000
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2 - 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 3 - 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 4 - 3000

    def test_sub_with_keys_differences_with_mixed_default_nodefault(
            self, smallts_withdefault, otherts):
        ts = smallts_withdefault - otherts
        assert not ts._has_default
        assert CURRENT + 1 * ONEHOUR not in ts.index
        assert ts[CURRENT + 2 * ONEHOUR] == 2 - 1000
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2 - 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 3 - 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 4 - 3000


class TestTictsMul:
    def test_simple_mul(self, smallts):
        ts = smallts * smallts
        assert list(ts.values()) == [val * val for val in smallts.values()]

    def test_simple_mul_one_float(self, smallts):
        ts = smallts * 1000.
        assert list(ts.values()) == [val * 1000. for val in smallts.values()]


class TestTictsDiv:
    def test_simple_div(self, smallts):
        smallts[CURRENT] = 1  # can't divide by zero
        ts = smallts / smallts
        assert list(ts.values()) == [val / val for val in smallts.values()]

    def test_simple_div_with_default_to_zero(self, smallts):
        smallts.default = 0
        smallts[CURRENT] = 1  # can't divide by zero
        ts = smallts / smallts
        assert not ts._has_default

    def test_simple_div_one_float(self, smallts):
        ts = smallts / 1000.
        assert list(ts.values()) == [val / 1000. for val in smallts.values()]


class TestTictsBoolean:
    def test_simple_le(self, smallts):
        result = smallts >= 5
        assert all([
            not val for val in result[CURRENT:CURRENT + 4 * ONEHOUR].values()
        ])
        assert all([
            val for val in result[CURRENT + 5 * ONEHOUR:CURRENT +
                                  9 * ONEHOUR].values()
        ])

    def test_simple_lt(self, smallts):
        result = smallts > 5
        assert all([
            not val for val in result[CURRENT:CURRENT + 5 * ONEHOUR].values()
        ])
        assert all([
            val for val in result[CURRENT + 6 * ONEHOUR:CURRENT +
                                  9 * ONEHOUR].values()
        ])

    def test_simple_eq(self, smallts):
        ts = smallts == smallts
        assert all(ts.values())


class TestTictsFloorCeil:
    def test_floor_on_float(self, smallts):
        ts = smallts.floor(2)
        assert all([value <= 2 for value in ts.values()])

    def test_floor_on_ts(self, smallts_withdefault, otherts_withdefault):
        ts = smallts_withdefault.floor(otherts_withdefault)
        assert ts[CURRENT + 1 * ONEHOUR] == 1
        assert ts[CURRENT + 2 * ONEHOUR] == 2
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2
        assert ts[CURRENT + 3 * ONEHOUR] == 3
        assert ts[CURRENT + 4 * ONEHOUR] == 4

    def test_ceil_on_float(self, smallts):
        ts = smallts.ceil(7)
        assert all([value >= 7 for value in ts.values()])

    def test_ceil_on_ts(self, smallts_withdefault, otherts_withdefault):
        ts = smallts_withdefault.ceil(otherts_withdefault)
        assert ts[CURRENT + ONEHOUR] == 900
        assert ts[CURRENT + 2 * ONEHOUR] == 1000
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 3000

    def test_ceil_on_ts_check_changing_default(self, smallts_withdefault,
                                               otherts_withdefault):
        otherts_withdefault.default = -10
        ts = smallts_withdefault.ceil(otherts_withdefault)

        assert ts[CURRENT + ONEHOUR] == 1
        assert ts[CURRENT + 2 * ONEHOUR] == 1000
        assert ts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 3000


class TestMaskUpdate:
    """
                0    1    2    3    4    5    6    7    8    9
                |    |    |    |    |    |    |    |    |    |    |
                ----------------------------------------------------------
    smallts     *    *    *    *    *    *    *    *    *    *
    otherts               *  *      *
    maskts                   T      F
    """

    def test_simple_mask_update_raises_on_type_of_other(self, smallts, maskts):
        with pytest.raises(TypeError) as err:
            smallts.mask_update("fakedct", maskts)

        assert "other should be of type TimeSeries, got" in str(err.value)

    def test_simple_mask_update_raises_if_not_boolean(self, smallts):
        with pytest.raises(TypeError) as err:
            smallts.mask_update(smallts, smallts)

        assert "The values of the mask should all be boolean" in str(err.value)

    def test_mask_update_with_empty_no_default_other_raises(
            self, smallts, emptyts, maskts):
        with pytest.raises(ValueError) as err:
            smallts.mask_update(emptyts, maskts)
        assert 'other is empty and has no default set' in str(err.value)

    def test_mask_update_with_empty_and_no_default_mask_raises(
            self, smallts, emptyts):
        with pytest.raises(ValueError) as err:
            smallts.mask_update(smallts, emptyts)
        assert 'mask is empty and has no default set' in str(err.value)

    def test_simple_mask_update(self, smallts, otherts, maskts):
        smallts_keys = smallts.index
        otherts_keys = otherts.index
        smallts.mask_update(otherts, maskts)

        assert set(smallts.index) == set(smallts_keys).union(set(otherts_keys))

        assert smallts[CURRENT + 2 * ONEHOUR] == 2
        assert smallts[CURRENT + 2 * ONEHOUR + HALFHOUR] == 2000
        assert smallts[CURRENT + 3 * ONEHOUR] == 2000
        assert smallts[CURRENT + 4 * ONEHOUR] == 4

    def test_mask_update_with_other_being_empty_with_default(
            self, smallts, emptyts, maskts):
        emptyts.default = -10.
        smallts.mask_update(emptyts, maskts)

        assert smallts[CURRENT + 2 * ONEHOUR] == 2
        assert smallts[CURRENT + 3 * ONEHOUR] == -10.
        assert smallts[CURRENT + 4 * ONEHOUR] == 4

    def test_mask_update_on_emptyts(self, smallts, emptyts, maskts):
        emptyts.mask_update(smallts, maskts)
        expected_ts = TimeSeries({CURRENT + 3 * ONEHOUR: 3})
        testing.assert_ts_equal(emptyts, expected_ts)



import pandas as pd
import pytest
from pandas.tseries.frequencies import to_offset

from ticts import TimeSeries, testing

from .conftest import CURRENT, HALFHOUR, ONEHOUR, ONEMIN


class TestToDataFrame:
    def test_to_dataframe(self, smallts):
        smallts.name = 'SuperTS'
        df = smallts.to_dataframe()
        assert isinstance(df, pd.DataFrame)
        assert df.columns.to_list() == ['SuperTS']
        assert df.index.freq == '1H'

    def test_it_works_on_unevenly_spaced(self, otherts):
        otherts.name = 'SuperTS'
        df = otherts.to_dataframe()
        assert isinstance(df, pd.DataFrame)
        assert df.columns.to_list() == ['SuperTS']
        assert df.index.freq is None


def test_from_df_to_ticst(smallts):
    df = smallts.to_dataframe()
    returned = df.to_ticts()
    testing.assert_ts_equal(smallts, returned)


def test_from_series_to_ticst(smallts):
    serie = smallts.to_dataframe()['value']
    returned = serie.to_ticts()
    testing.assert_ts_equal(smallts, returned)


class TestTimeSeriesSample:
    def test_it_raises_when_both_freq_and_index_are_none(self, smallts):
        with pytest.raises(Exception) as err:
            smallts.sample()
        assert "Both are None" in str(err.value)

    def test_simple_sample_on_default(self, smallts):
        ts = smallts.sample(HALFHOUR)
        expected_dict = {
            **smallts.data,
            **{key + HALFHOUR: smallts[key]
               for key in smallts.index[:-1]}
        }
        expected_ts = TimeSeries(expected_dict, default=smallts.default)
        testing.assert_ts_equal(expected_ts, ts)

    def test_simple_by_passing_an_index(self, smallts):
        index = pd.date_range(
            start=smallts.lower_bound,
            end=smallts.upper_bound + 10 * ONEHOUR,
            freq='3H',
            closed='left')
        ts = smallts.sample(index=index)
        assert list(ts.index) == index.to_list()

    def test_simple_sample_with_start_being_string(self, smallts):
        start = '2019-01-01T09:00:00'
        ts = smallts.sample(HALFHOUR, start=start)
        expected_ts = TimeSeries({start: 9})
        testing.assert_ts_equal(expected_ts, ts)

    def test_sample_out_of_left_bound_with_no_default_permissive_shorten_interval(
            self, smallts):
        ts = smallts.sample(
            freq=HALFHOUR, start=CURRENT - ONEHOUR, end=CURRENT + ONEHOUR)
        expected_ts = TimeSeries({CURRENT: 0, CURRENT + HALFHOUR: 0})
        testing.assert_ts_equal(expected_ts, ts)

    def test_sample_on_empty_return_empty_df(self, emptyts):
        assert emptyts.sample(freq=ONEHOUR).empty

    def test_sample_with_start_out_of_left_bounds_with_default(
            self, smallts_withdefault):
        ts = smallts_withdefault.sample(freq=HALFHOUR, start=CURRENT - ONEHOUR)
        default = smallts_withdefault.default
        assert ts[CURRENT - ONEHOUR] == default
        assert ts[CURRENT - HALFHOUR] == default
        assert ts[CURRENT] == 0
        assert CURRENT + HALFHOUR in ts.index

    def test_sample_with_start_out_of_left_bounds(self, smallts):
        ts = smallts.sample(freq=HALFHOUR, start=CURRENT - ONEHOUR)
        CURRENT - ONEHOUR not in ts.index
        CURRENT - HALFHOUR not in ts.index
        assert ts[CURRENT] == 0
        assert CURRENT + HALFHOUR in ts.index

    @pytest.mark.parametrize('freq', [
        ONEMIN,
        10 * ONEMIN,
        '10Min',
        pd.Timedelta(seconds=10),
        ONEHOUR,
        '1H',
    ])
    def test_sample_on_freq_with_end_and_start_in_bound(self, otherts, freq):
        start = CURRENT + 2 * ONEHOUR + HALFHOUR
        end = CURRENT + 3 * ONEHOUR + HALFHOUR
        ts = otherts.sample(freq=freq, start=start, end=end)
        freq = pd.Timedelta(freq)
        expected = [start + i * freq for i in range(int((end - start) / freq))]
        assert list(ts.index) == expected

    @pytest.mark.parametrize('freq', ['MS', 'M', 'B'])
    def test_sample_on_non_fixed_frequency(self, smallts, freq):
        ts = smallts.sample(freq)
        for start, end in ts.iterintervals():
            to_offset(end - start) == to_offset(freq)

    def test_sample_on_non_fixed_frequency_with_values(self, smallts):
        freq = to_offset('MS')
        ts = smallts.sample(freq)
        expected_index = [smallts.lower_bound, smallts.lower_bound + freq]
        list(ts.index) == expected_index
        expected_values = [
            smallts[smallts.lower_bound], smallts[smallts.lower_bound + freq]
        ]
        list(ts.values()) == expected_values



from datetime import timedelta

from ticts import TimeSeries
from ticts.utils import timestamp_converter

onehour = timedelta(hours=1)
onemin = timedelta(minutes=1)
dt1 = timestamp_converter('2019-01-01')
dt2 = timestamp_converter('2019-01-02')

smalldct = {
    dt1 + onehour: 1,
    dt1 + 3 * onehour: 3,
    dt1 + 6 * onehour: 10,
}

smallts = TimeSeries(smalldct)

otherdct = {
    dt1 + 2 * onehour: 2,
    dt1 + 3 * onehour: 3,
    dt1 + 5 * onehour: 5,
}

otherts = TimeSeries(otherdct, default=0)



import logging

from .utils import MINTS, NO_DEFAULT, operation_factory

logger = logging.getLogger(__name__)


def _get_keys_for_operation(ts1, ts2, *args):
    all_ts = [ts1, ts2, *args]
    for ts in all_ts:
        if not ts.__class__.__name__ == 'TimeSeries':
            raise TypeError("{} is not of type TimeSeries".format(ts))

    all_keys = set.union(*[set(ts.index) for ts in all_ts])

    lower_bound = MINTS
    for ts in all_ts:
        if not ts._has_default:
            lower_bound = max(lower_bound, ts.lower_bound)

    return [key for key in all_keys if key >= lower_bound]


class TictsOperationMixin:
    def _operate(self, other, operator):
        if isinstance(other, self.__class__):
            return self._operate_on_ts(other, operator)
        else:
            return self._operate_on_one_value(other, operator)

    def _operate_on_ts(self, other, operator):
        if not isinstance(other, self.__class__):
            raise TypeError

        all_keys = set(self.index).union(set(other.index))

        default = NO_DEFAULT
        if self._has_default and other._has_default:
            try:
                default = operator(self.default, other.default)
            except ZeroDivisionError:
                default = NO_DEFAULT
                msg = ("The TimeSeries has 0. as default value."
                       " Can't compute the resulting default.")
                logger.warning(msg)

        all_keys = _get_keys_for_operation(self, other)

        ts = self.__class__(default=default)
        for key in all_keys:
            ts[key] = operator(self[key], other[key])

        return ts

    def _operate_on_one_value(self, value, operator):
        sample_value = self.values()[0]
        try:
            operator(sample_value, value)
        except Exception:
            msg = "Can't apply {} on {} with {}."
            raise TypeError(
                msg.format(operator.__name__, sample_value.__class__.__name__,
                           value.__class__.__name__))

        default = None
        if self._has_default:
            default = operator(self.default, value)

        ts = self.__class__(default=default)
        for key in self.index:
            ts[key] = operator(self[key], value)

        return ts

    __add__ = operation_factory('__add__')
    __radd__ = operation_factory('__add__')
    __sub__ = operation_factory('__sub__')

    __mul__ = operation_factory('__mul__')
    __truediv__ = operation_factory('__truediv__')
    __floordiv__ = operation_factory('__floordiv__')

    __abs__ = operation_factory('__abs__')

    __lt__ = operation_factory('__lt__')
    __le__ = operation_factory('__le__')
    __gt__ = operation_factory('__gt__')
    __ge__ = operation_factory('__ge__')
    __eq__ = operation_factory('__eq__')

    __or__ = operation_factory('__or__')
    __xor__ = operation_factory('__xor__')
    __and__ = operation_factory('__and__')

    __inv__ = operation_factory('__inv__')
    __not__ = operation_factory('__not__')

    def floor(self, other):
        """Floor your timeseries, applying a min key by key.

        Args:
            other (TimeSeries or numeric): values to floor on.

        Returns:
            TimeSeries floored
        """
        return self._operate(other, min)

    def ceil(self, other):
        """Ceil your timeseries, applying a max key by key.

        Args:
            other (TimeSeries or numeric): values to ceil on.

        Returns:
            TimeSeries ceiled
        """
        return self._operate(other, max)

    def mask_update(self, other, mask):
        """Update your timeseries with another one in regards of a mask.

        Args:
            other (TimeSeries): values taken to update.
            mask (TimeSeries): timeseries with boolean values.

        Returns:
            TimeSeries
        """
        # Type checks
        if not isinstance(other, self.__class__):
            msg = 'other should be of type TimeSeries, got {}'
            raise TypeError(msg.format(type(other)))

        if not all([isinstance(value, bool) for value in mask.values()]):
            msg = 'The values of the mask should all be boolean.'
            raise TypeError(msg)

        # Empty ts checks
        if mask.empty and not mask._has_default:
            msg = "mask is empty and has no default set"
            raise ValueError(msg)

        if other.empty and not other._has_default:
            msg = "other is empty and has no default set"
            raise ValueError(msg)

        all_keys = _get_keys_for_operation(self, other)

        if not mask._has_default:
            all_keys = [key for key in all_keys if key >= mask.lower_bound]

        for key in all_keys:
            if mask[key]:
                self[key] = other[key]
                self[key] = other[key]



import json

from ticts import TimeSeries, testing


class TestJSON:
    def test_serealize_iso(self, smallts):
        expected = {
            'data': {key.isoformat(): value
                     for key, value in smallts.items()},
            'default': 'no_default',
            'name': 'value'
        }
        assert expected == smallts.serealize(date_format='iso')

    def test_serealize_epoch(self, smallts):
        expected = {
            'data': {key.value: value
                     for key, value in smallts.items()},
            'default': 'no_default',
            'name': 'value'
        }
        assert expected == smallts.serealize()

    def test_to_json(self, smallts, tmpdir):
        path = tmpdir.join('test.json')
        smallts.to_json(path)
        expected = smallts.serealize()
        expected['data'] = {
            str(key): val
            for key, val in expected['data'].items()
        }
        returned = json.load(open(path, "r"))
        assert returned == expected

    def test_it_returns_timeseries_from_json(self, smallts, tmpdir):
        path = tmpdir.join('test.json')
        smallts.to_json(path)
        ts_read = TimeSeries.from_json(path)
        testing.assert_ts_equal(smallts, ts_read)



import json
from pathlib import Path

import pandas as pd

from .utils import NO_DEFAULT


class TictsIOMixin:
    def serealize(self, date_format='epoch'):
        if date_format.lower() == 'epoch':
            keys = [key.value for key in self.index]
        elif date_format.lower() in ['iso', 'isoformat']:
            keys = [key.isoformat() for key in self.index]
        else:
            msg = "Date serealize with date_format equal to '{}' is not implemented"
            raise NotImplementedError(msg.format(date_format))

        return {
            'data': {key: val
                     for key, val in zip(keys, self.values())},
            'default': self.default
            if self.default != NO_DEFAULT else 'no_default',
            'name': self.name,
        }

    def to_json(self, path_or_buf, date_format='epoch', compression='infer'):
        path_or_buf = pd.io.common._stringify_path(path_or_buf)

        s = json.dumps(self.serealize())

        if isinstance(path_or_buf, str):
            fh, handles = pd.io.common._get_handle(
                path_or_buf, 'w', compression=compression)
            try:
                fh.write(s)
            finally:
                fh.close()
        elif path_or_buf is None:
            return s
        else:
            path_or_buf.write(s)

    @classmethod
    def from_json(cls, path):
        if not hasattr(path, 'read'):
            path = Path(path)
            if not path.exists():
                raise Exception("'{}' does not exists.".format(path))
            path = open(path, mode="r")

        content = json.load(path)
        return cls(**content)



from copy import copy, deepcopy
from datetime import datetime
from unittest import mock

import pandas as pd
import pytest

from ticts import TimeSeries, testing
from ticts.utils import MAXTS, MINTS, timestamp_converter

from .conftest import CURRENT, HALFHOUR, ONEHOUR, ONEMIN


class TestTimeSeriesInit:
    def test_with_dict(self, smalldict):
        ts = TimeSeries(smalldict, default=10)
        assert ts[CURRENT + ONEHOUR] == 1
        assert ts.default == 10

    def test_with_dict_keys_being_strings(self):
        dct = {
            '2019-01-01': 1,
            '2019-02-01': 2,
            timestamp_converter('2019-03-01'): 3,
        }
        ts = TimeSeries(dct, default=10)
        expected = {
            timestamp_converter(key): value
            for key, value in dct.items()
        }
        assert ts.items() == expected.items()
        assert ts.default == 10

    def test_with_data_as_tuple(self):
        mytuple = (
            (CURRENT, 0),
            ('2019-02-01', 1),
        )
        ts = TimeSeries(mytuple, default=10)
        assert ts[CURRENT] == 0
        assert ts['2019-02-01'] == 1
        assert len(ts) == 2
        assert ts.default == 10

    def test_with_data_as_pandas_series(self, smalldict):
        serie = pd.Series(data=smalldict, name='SomeName')
        ts = TimeSeries(serie)
        assert ts.name == 'SomeName'
        assert ts[CURRENT] == 0
        assert len(ts) == len(serie)

    def test_with_data_as_dataframe(self, smalldict):
        df = pd.DataFrame(
            data={'SomeName': list(smalldict.values())},
            index=smalldict.keys())
        ts = TimeSeries(df, default=10)
        assert ts[CURRENT] == 0
        assert ts.name == 'SomeName'
        assert len(ts) == df.shape[0]
        assert ts.default == 10

    def test_with_data_as_dataframe_raises_when_several_columns(self):
        df = pd.DataFrame(columns=['Too', 'Many', 'Columns'])
        with pytest.raises(Exception) as err:
            TimeSeries(df)

        expected = "Can't convert a DataFrame with several columns"
        assert expected in str(err.value)

    def test_with_data_as_ticts_timeserie(self, smallts):
        smallts.default = 1000
        smallts.name = 'SomeName'
        testing.assert_ts_equal(smallts, TimeSeries(smallts))

    def test_with_data_as_ticts_timeserie_and_default_given(self, smallts):
        smallts.default = 1000
        smallts.name = 'SomeName'
        newts = TimeSeries(smallts, default=1.)
        testing.assert_ts_equal(smallts, newts, check_default=False)
        assert newts.default == 1.

    def test_with_data_as_ticts_timeserie_and_name_given(self, smallts):
        smallts.default = 1000
        smallts.name = 'SomeName'
        newts = TimeSeries(smallts, name='SomeOtherName')
        testing.assert_ts_equal(smallts, newts, check_name=False)
        assert newts.name == 'SomeOtherName'


class TestTimeSeriesSetItem:
    def test_simple_setitem(self, smallts):
        smallts[CURRENT] = 1000
        assert smallts[CURRENT] == 1000

    def test_consecutive_setitem(self, smallts):
        smallts[CURRENT] = 1000
        first_time = TimeSeries(smallts)
        smallts[CURRENT] = 1000
        testing.assert_ts_equal(first_time, smallts)

    @mock.patch("ticts.TimeSeries.set_interval")
    def test_setitem_on_slice_calls_set_interval(self, set_interval, smallts):
        smallts[CURRENT:CURRENT + 2 * ONEHOUR] = 1000
        assert set_interval.call_count == 1


class TestTictsMagicMixin:
    # Copy / Deepcopy

    def test_copy(self, smallts):
        copied = copy(smallts)
        testing.assert_ts_equal(copied, smallts)

    def test_copy_with_default(self, smallts_withdefault):
        copied = copy(smallts_withdefault)
        testing.assert_ts_equal(copied, smallts_withdefault)

    def test_deepcopy(self, smallts):
        deepcopied = deepcopy(smallts)
        testing.assert_ts_equal(deepcopied, smallts)

    def test_deepcopy_with_default(self, smallts_withdefault):
        deepcopied = deepcopy(smallts_withdefault)
        testing.assert_ts_equal(deepcopied, smallts_withdefault)

    # Repr

    def test_repr_on_otherts(self, otherts):
        assert repr(otherts).count('\n') == 3

    def test_repr_on_large(self):
        dct = dict()
        for i in range(100):
            dct[CURRENT + i * ONEHOUR] = i
        ts = TimeSeries(dct)
        assert repr(ts).count('\n') == 11

    def test_repr_with_default_and_name(self, smallts_withdefault):
        smallts_withdefault.name = 'SuperCool'
        assert 'default=' in repr(smallts_withdefault)
        assert 'name=' in repr(smallts_withdefault)

    # Accessors

    def test_values(self, smallts):
        values = smallts.values()
        series_values = [v for k, v in smallts.items()]

        assert len(values) == len(series_values)

    def test_keys(self, smallts):
        keys = smallts.keys()
        assert len(keys) == len(smallts.index)

    # Delitem

    def test_delitem(self, smallts):
        del smallts[CURRENT]
        assert CURRENT not in smallts.index

    # Update

    def test_update(self, smallts):
        og_ts = deepcopy(smallts)
        last_index = smallts.index[-1]
        new_index = last_index + pd.Timedelta('2 days')
        new_value = 10

        # Updating with TimeSeries
        other = TimeSeries({new_index: new_value, last_index: new_value})
        og_ts.update(other)

        assert og_ts[new_index] == new_value
        assert og_ts[last_index] == new_value

        # Updating with tuple
        og_ts = deepcopy(smallts)
        other_tuple = tuple(other.items())
        og_ts.update(other_tuple)

        assert og_ts[new_index] == new_value

        # Updating with *args
        og_ts = deepcopy(smallts)
        other_new_index = new_index + pd.Timedelta('2 hours')
        other_new = TimeSeries({other_new_index: new_value})
        og_ts.update([other, other_new])

        assert og_ts[new_index] == new_value
        assert og_ts[other_new_index] == new_value

        # Updating with dict
        og_ts = deepcopy(smallts)
        og_ts.update({new_index: new_value})

        assert og_ts[new_index] == new_value

        # Updating with **kwargs
        og_ts = deepcopy(smallts)
        with pytest.raises(TypeError):
            og_ts.update(**{new_index: new_value})


class TestTimeSeriesDefault:
    @pytest.mark.parametrize('default', [None, 0, False])
    def test_it_has_default(self, default):
        ts = TimeSeries(default=default)
        assert ts._has_default

    def test_it_has_no_default(self):
        ts = TimeSeries()
        assert not ts._has_default


class TestTimeSeriesEqual:
    def test_equals(self, smallts_withdefault):
        assert smallts_withdefault.equals(smallts_withdefault) is True

    def test_not_equals(self, smallts, otherts):
        assert smallts.equals(otherts) is False

    def test_not_equals_due_to_default(self, smallts, smallts_withdefault):
        assert smallts_withdefault.equals(smallts) is False

    def test_equals_without_default(self, smallts, smallts_withdefault):
        assert smallts_withdefault.equals(smallts, check_default=False) is True


class TestTimeSeriesBoundProperties:
    def test_lower_bound(self, smallts):
        assert smallts.lower_bound == smallts.index[0]

    def test_upper_bound(self, smallts):
        assert smallts.upper_bound == smallts.index[-1]

    def test_lower_bound_on_empty(self, emptyts):
        assert emptyts.lower_bound == MINTS

    def test_upper_bound_on_empty(self, emptyts):
        assert emptyts.upper_bound == MAXTS


def test_timeseries_compact(smallts):
    smallts[CURRENT + ONEMIN] = 0
    assert (CURRENT + ONEMIN) not in smallts.compact().index


class TestTimeSeriesGetitem:
    available_interpolate = ['previous', 'linear']

    # tests on corner cases

    def test_get_item_out_of_left_bound_with_default_zero(self):
        mytuple = ((CURRENT, 0), (CURRENT + ONEHOUR, 1))
        ts = TimeSeries(mytuple, default=0)
        assert ts[CURRENT + ONEHOUR] == 1
        assert ts[CURRENT - ONEHOUR] == 0

    def test_getitem_out_of_left_bound_with_no_default_and_permissive_return_None(
            self, smallts):
        assert smallts[CURRENT - ONEMIN] is None

    def test_getitem_out_of_left_bound_with_no_default_and_not_permissive_raises(
            self, smallts):
        smallts.permissive = False
        with pytest.raises(KeyError) as err:
            smallts[CURRENT - ONEMIN]
        assert 'default attribute is not set' in str(err.value)

    def test_getitem_using_str(self, smallts):
        smallts['2019-01-01'] == 0

    @pytest.mark.parametrize('interpolate', available_interpolate)
    def test_getitem_out_of_left_bound_with_default_return_default(
            self, smallts_withdefault, interpolate):
        value = smallts_withdefault[CURRENT - ONEMIN, interpolate]
        assert value == smallts_withdefault.default

    @pytest.mark.parametrize('interpolate', available_interpolate)
    def test_getitem_on_empty_when_no_default_not_permissive_raises(
            self, emptyts, interpolate):
        emptyts.permissive = False
        with pytest.raises(KeyError) as err:
            emptyts[CURRENT - ONEMIN, interpolate]

        expected = "default attribute is not set and timeseries is empty"
        assert expected in str(err.value)

    @pytest.mark.parametrize('interpolate', available_interpolate)
    def test_getitem_on_empty_when_no_default_return_None(
            self, emptyts, interpolate):
        emptyts[CURRENT - ONEMIN, interpolate] is None

    # tests on '_get_previous'

    @mock.patch("ticts.TimeSeries._get_previous")
    def test_get_on_previous_is_default_interpolate(self, _get_previous,
                                                    smallts):
        smallts[CURRENT + ONEMIN]
        assert _get_previous.call_count == 1

    def test_get_on_previous(self, smallts):
        assert smallts[CURRENT + ONEMIN] == 0

    def test_get_on_previous_out_of_right_bound(self, smallts):
        assert smallts[CURRENT + 10 * ONEHOUR] == 9

    def test_get_on_previous_out_of_left_bound_no_default(self, smallts):
        assert smallts[CURRENT - ONEHOUR] is None

    def test_get_on_previous_out_of_left_bound_with_default(
            self, smallts_withdefault):
        assert smallts_withdefault[CURRENT -
                                   ONEHOUR] == smallts_withdefault.default

    # tests on '_get_linear_interpolate'

    @pytest.mark.parametrize('time_idx, expected', [
        (CURRENT + HALFHOUR, 0.5),
        (CURRENT + ONEHOUR + HALFHOUR, 1.5),
        (CURRENT + 10 * ONEMIN, 0 + (1 - 0) * (10 * ONEMIN / ONEHOUR)),
    ])
    def test_get_linear_interpolate(self, smallts, time_idx, expected):
        assert smallts[time_idx, 'linear'] == expected

    def test_get_linear_interpolate_out_of_right_bound(self, smallts):
        assert smallts[CURRENT + 10 * ONEHOUR, 'linear'] == 9

    def test_get_linear_interpolate_out_of_left_bound_no_default(
            self, smallts):
        assert smallts[CURRENT - ONEHOUR, 'linear'] is None

    def test_get_linear_interpolate_out_of_left_bound_with_default(
            self, smallts_withdefault):
        assert smallts_withdefault[CURRENT - ONEHOUR,
                                   'linear'] == smallts_withdefault.default

    # test on 'slice'

    def test_get_on_slice_exclude_upper_bound_include_lower_bound(
            self, smallts):
        start = CURRENT
        end = CURRENT + 1 * ONEHOUR
        sliced_ts = smallts[start:end + ONEHOUR]

        data = {start: 0, end: 1}
        expected_ts = TimeSeries(data, default=smallts.default)
        testing.assert_ts_equal(sliced_ts, expected_ts)

    def test_get_on_slice_add_back_previous_value_if_start_not_in_keys(
            self, smallts):
        start = CURRENT + HALFHOUR
        end = CURRENT + ONEHOUR
        sliced_ts = smallts[start:end + ONEHOUR]

        data = {start: 0, end: 1}
        expected_ts = TimeSeries(data, default=smallts.default)
        testing.assert_ts_equal(sliced_ts, expected_ts)

    def test_get_on_slice_entirely_out_of_bounds_on_left_side(self, smallts):
        assert smallts[CURRENT - 2 * ONEHOUR:CURRENT - 1 * ONEHOUR].empty

    def test_get_on_slice_out_of_bounds_left_side(self, smallts):
        start = CURRENT - 2 * ONEHOUR
        end = CURRENT
        data = {CURRENT: 0}
        sliced_ts = smallts[start:end + 1 * ONEHOUR]
        expected_ts = TimeSeries(data, default=smallts.default)
        testing.assert_ts_equal(sliced_ts, expected_ts)

    def test_get_on_slice_entirely_out_of_bounds_on_right_side(self, smallts):
        start = CURRENT + 10 * ONEHOUR
        end = CURRENT + 12 * ONEHOUR
        sliced_ts = smallts[start:end]
        expected_ts = TimeSeries({start: smallts[start]})
        testing.assert_ts_equal(sliced_ts, expected_ts)

    def test_get_on_slice_out_of_bounds_on_right_side(self, smallts):
        sliced_ts = smallts[CURRENT + 9 * ONEHOUR:CURRENT + 12 * ONEHOUR]
        expected_dct = {CURRENT + 9 * ONEHOUR: 9}
        expected_ts = TimeSeries(expected_dct, default=smallts.default)
        testing.assert_ts_equal(sliced_ts, expected_ts)


class TestTimeSeriesSetInterval:
    def test_set_interval_when_no_default_raises(self, smallts):
        with pytest.raises(NotImplementedError):
            smallts.set_interval(CURRENT, CURRENT + ONEHOUR, 1000)

    def test_single_set_interval_end_on_last_key(self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT + ONEHOUR,
                                         CURRENT + 9 * ONEHOUR, 1000)
        expected_keys = [CURRENT, CURRENT + ONEHOUR, CURRENT + 9 * ONEHOUR]
        assert list(smallts_withdefault.index) == expected_keys
        assert smallts_withdefault[CURRENT + ONEHOUR] == 1000
        assert smallts_withdefault[CURRENT + 9 * ONEHOUR] == 9.

    def test_single_set_interval_start_on_first_key(self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT, CURRENT + 9 * ONEHOUR, 1000)
        expected_keys = [CURRENT, CURRENT + 9 * ONEHOUR]
        assert list(smallts_withdefault.index) == expected_keys
        assert smallts_withdefault[CURRENT] == 1000
        assert smallts_withdefault[CURRENT + 9 * ONEHOUR] == 9.

    @pytest.mark.parametrize('start, end', [
        (CURRENT + ONEHOUR, CURRENT + 10 * ONEHOUR),
        (CURRENT + 4 * ONEHOUR, CURRENT + 11 * ONEHOUR),
    ])
    def test_single_set_interval_end_over_last_key_sets_to_last_value(
            self, smallts_withdefault, start, end):
        last_key = smallts_withdefault.index[-1]
        last_val = smallts_withdefault[last_key]

        smallts_withdefault.set_interval(start, end, 1000)
        keys_before_start = []
        for key in smallts_withdefault.index:
            if key < start:
                keys_before_start.append(key)

        expected_keys = [*keys_before_start, start, end]
        assert list(smallts_withdefault.index) == expected_keys
        assert smallts_withdefault[end] == last_val

    def test_single_set_interval_when_start_higher_than_upper_bound(
            self, smallts_withdefault):
        start = CURRENT + 11 * ONEHOUR
        end = CURRENT + 13 * ONEHOUR
        smallts_withdefault.set_interval(start, end, 1000)

        assert CURRENT + 10 * ONEHOUR not in smallts_withdefault.index
        assert smallts_withdefault[CURRENT + 10 * ONEHOUR] == 9

        assert smallts_withdefault[start] == 1000
        assert smallts_withdefault[end] == 9

    def test_single_set_interval_start_before_first_key(
            self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT - ONEHOUR,
                                         CURRENT + 9 * ONEHOUR, 1000)
        expected_keys = [CURRENT - ONEHOUR, CURRENT + 9 * ONEHOUR]
        assert list(smallts_withdefault.index) == expected_keys
        assert smallts_withdefault[CURRENT - 1 * ONEHOUR] == 1000

    def test_single_set_interval_on_bounds_not_being_keys(
            self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT + ONEMIN,
                                         CURRENT + 9 * ONEHOUR, 1000)
        expected_keys = [CURRENT, CURRENT + ONEMIN, CURRENT + 9 * ONEHOUR]
        assert list(smallts_withdefault.index) == expected_keys
        assert smallts_withdefault[CURRENT + ONEMIN] == 1000

    def test_set_interval_on_empty(self, emptyts):
        emptyts.default = 10
        emptyts.set_interval(CURRENT, CURRENT + ONEHOUR, 1)
        assert emptyts[CURRENT] == 1
        assert emptyts[CURRENT + ONEHOUR] == 10
        assert len(emptyts.index) == 2

    def test_set_interval_on_empty_with_default(self, emptyts_withdefault):
        emptyts_withdefault.set_interval(CURRENT, CURRENT + ONEHOUR, 1)
        assert emptyts_withdefault[CURRENT] == 1
        assert emptyts_withdefault[CURRENT +
                                   ONEHOUR] == emptyts_withdefault.default
        len(emptyts_withdefault.index) == 2

    @pytest.mark.parametrize('start, end', (
        (CURRENT, CURRENT + 9 * ONEHOUR),
        (CURRENT + 5 * ONEHOUR, CURRENT + 12 * ONEHOUR),
        (datetime(2019, 1, 1) + 5 * ONEHOUR,
         datetime(2019, 1, 1) + 12 * ONEHOUR),
        ('2019-01-01T05:00:00', '2019-01-01T12:00:00+02:00'),
    ))
    def test_same_consecutive_set_interval(self, smallts_withdefault, start,
                                           end):
        smallts_withdefault.set_interval(start, end, 1000)
        first_time = deepcopy(smallts_withdefault)

        for _ in range(10):
            smallts_withdefault.set_interval(start, end, 1000)

        testing.assert_ts_equal(first_time, smallts_withdefault)

    def test_consecutive_set_interval_on_empty_with_default(self, emptyts):
        emptyts.default = 10
        emptyts.set_interval(CURRENT, CURRENT + 2 * ONEHOUR, 10)
        emptyts.set_interval(CURRENT, CURRENT + ONEHOUR, 0)
        emptyts.set_interval(CURRENT + ONEHOUR, CURRENT + 2 * ONEHOUR, 1)
        emptyts.set_interval(CURRENT, CURRENT + 2 * ONEHOUR, 3)
        assert emptyts[CURRENT] == 3
        assert emptyts[CURRENT + 2 * ONEHOUR] == 10
        assert list(emptyts.index) == [CURRENT, CURRENT + 2 * ONEHOUR]

    def test_set_interval_when_no_keys_to_delete_with_default(self, emptyts):
        emptyts.default = 1000
        emptyts.set_interval(CURRENT, CURRENT + 1 * ONEHOUR, 0)
        emptyts.set_interval(CURRENT + 3 * ONEHOUR, CURRENT + 4 * ONEHOUR, 3)
        emptyts.set_interval(CURRENT + 1 * ONEHOUR + HALFHOUR,
                             CURRENT + 3 * ONEHOUR, 10)

        expected_dct = {
            CURRENT: 0,
            CURRENT + 1 * ONEHOUR: 1000.,
            CURRENT + 1 * ONEHOUR + HALFHOUR: 10,
            CURRENT + 3 * ONEHOUR: 3,
            CURRENT + 4 * ONEHOUR: 1000
        }
        assert list(emptyts.index) == list(expected_dct.keys())

        for key in expected_dct:
            assert emptyts[key] == expected_dct[key]


class TestIterIntervals:
    def test_simple_iterintervals(self, smallts):
        iterator = smallts.iterintervals()
        result = [(start, end) for start, end in iterator]
        expected = [(CURRENT + i * ONEHOUR, CURRENT + (i + 1) * ONEHOUR)
                    for i in range(9)]
        assert sorted(result) == sorted(expected)

    def test_iterintervals_with_end(self, smallts):
        end = CURRENT + 3 * ONEHOUR + HALFHOUR
        iterator = smallts.iterintervals(end)
        result = [(start, end) for start, end in iterator]
        expected = [(CURRENT + i * ONEHOUR, CURRENT + (i + 1) * ONEHOUR)
                    for i in range(3)]
        expected.append((CURRENT + 3 * ONEHOUR, end))
        assert sorted(result) == sorted(expected)


class TestTzConvert:
    def test_simple_tz_convert(self, smallts):
        ts = smallts.tz_convert('CET')
        assert ts.tz == 'CET'
        assert smallts.tz == 'UTC'


def test_chain_operations_keep_meta_keys(smallts_withdefault):
    ts = smallts_withdefault
    otherts = ts.compact().sample('1T')
    for attr_name in ts._meta_keys:
        assert getattr(ts, attr_name) == getattr(otherts, attr_name)



import logging
from copy import deepcopy

import pandas as pd
import pytz
from sortedcontainers import SortedDict, SortedList

from .io import TictsIOMixin
from .iplot import TictsPlot
from .operation import TictsOperationMixin
from .pandas_mixin import PandasMixin
from .utils import MAXTS, MINTS, NO_DEFAULT, timestamp_converter

logger = logging.getLogger(__name__)

DEFAULT_NAME = 'value'


def _process_args(data, tz):
    if data is None:
        data = []
    if isinstance(data, (list, tuple, set)) and len(data) == 1:
        data = data[0]

    if isinstance(data, pd.DataFrame):
        # We should already have checked len(df.columns) == 1
        data = data.to_dict()[data.columns[0]]
    elif isinstance(data, pd.Series):
        data = data.to_dict()

    if hasattr(data, 'items'):
        data = data.items()

    return ((timestamp_converter(k, tz), v) for k, v in data)


class TictsMagicMixin:
    def __copy__(self):
        return self.__class__(self)

    def __deepcopy__(self, memo):
        return self.__class__(
            data=deepcopy(self.data), **self._kwargs_special_keys)

    def __repr__(self):
        header = "<TimeSeries>"

        meta = []
        if self._has_default:
            meta.append('default={}'.format(self.default))
        if self.name != DEFAULT_NAME:
            meta.append("name='{}'".format(self.name))

        if meta:
            header = "{} ({})".format(header, ' | '.join(meta))

        def generate_content(keys):
            return '\n'.join(
                ["{}: {},".format(key.isoformat(), self[key]) for key in keys])

        if len(self) < 10:
            content = generate_content(self.index)
        else:
            content_head = generate_content(self.index[:5])
            content_tail = generate_content(self.index[-5:])
            content = "{}\n[...]\n{}".format(content_head, content_tail)

        return "{}\n{}".format(header, content)

    # Methods redirecting to SortedDict data attribute method
    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return self.data.__iter__()

    def __delitem__(self, key):
        del self.data[key]

    def items(self):
        return self.data.items()

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def update(self, *args, **kwargs):
        """Due to string-requirement of the keyword, kwargs are not supported.

        They are present only to duplicate the parent's signature.
        """
        new_args = []

        for arg in args:
            if isinstance(arg, TimeSeries):
                new_args.append(arg.items())

            elif isinstance(arg, (list, tuple, set)):
                new_arg = []

                for a in arg:
                    if isinstance(a, TimeSeries):
                        new_arg.extend(list(a.items()))
                    else:
                        new_arg.append(a)

                new_args.append(new_arg)

            else:
                new_args = args

        self.data.update(*new_args, **kwargs)


class TimeSeries(TictsMagicMixin, TictsOperationMixin, PandasMixin,
                 TictsIOMixin, TictsPlot):
    """ TimeSeries object.

    Args:
        default: The default value of timeseries.
        permissive (bool): Whether to allow accessing non-existing values or not.
            If is True, getting non existing item returns None.
            If is False, getting non existing item raises.
    """
    _default_interpolate = "previous"

    _meta_keys = ('default', 'name', 'permissive')

    @property
    def index(self):
        return self.data.keys()

    @property
    def lower_bound(self):
        """Return the lower bound time index."""
        if self.empty:
            return MINTS
        return self.index[0]

    @property
    def upper_bound(self):
        """Return the upper bound time index."""
        if self.empty:
            return MAXTS
        return self.index[-1]

    @property
    def _has_default(self):
        return self.default != NO_DEFAULT

    @property
    def _kwargs_special_keys(self):
        kwargs = {}
        for attr_name in self._meta_keys:
            kwargs[attr_name] = getattr(self, attr_name)
        return kwargs

    @property
    def empty(self):
        """Return whether the TimeSeries is empty or not."""
        return len(self) == 0

    def __init__(self,
                 data=None,
                 default=NO_DEFAULT,
                 name=DEFAULT_NAME,
                 permissive=True,
                 tz='UTC'):
        """"""
        if isinstance(data, self.__class__):
            for attr in ('data', *self._meta_keys):
                setattr(self, attr, getattr(data, attr))

            # Only set 'default' and 'name' if is different from default
            if default != NO_DEFAULT:
                setattr(self, 'default', default)
            if name != DEFAULT_NAME:
                setattr(self, 'name', name)
            return

        if hasattr(default, 'lower') and default.lower() == 'no_default':
            # 'no_default' as string is used at JSON serealization time
            self.default = NO_DEFAULT
        else:
            self.default = default

        self.name = name
        self.permissive = permissive

        # Overwrite the name if data is an instance of pd.DataFrame or pd.Series
        if isinstance(data, pd.DataFrame):
            if len(data.columns) != 1:
                msg = ("Can't convert a DataFrame with several columns into "
                       "one timeseries: {}.")
                raise ValueError(msg.format(data.columns))
            self.name = data.columns[0]

        elif isinstance(data, pd.Series):
            self.name = data.name

        try:
            tz = pytz.timezone(tz)
        except pytz.UnknownTimeZoneError:
            raise ValueError('{} is not a valid timezone'.format(tz))

        # SortedDict.__init__ does not use the __setitem__
        # Hence we got to parse datetime keys ourselves.
        # SortedDict use the first arg given and check if is a callable
        # in case you want to give your custom sorting function.
        self.data = SortedDict(None, _process_args(data, tz))

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            return self.set_interval(key.start, key.stop, value)
        if key in self._meta_keys:
            super().__setitem__(key, value)
        else:
            key = timestamp_converter(key, self.tz)
            self.data[key] = value

    def __getitem__(self, key):
        """Get the value of the time series, even in-between measured values by interpolation.
        Args:
            key (datetime): datetime index
            interpolate (str): interpolate operator among ["previous", "linear"]
        """

        interpolate = self._default_interpolate

        if isinstance(key, tuple):
            if len(key) == 2:
                key, interpolate = key
            elif len(key) > 2:
                raise KeyError

        if isinstance(key, slice):
            return self.slice(key.start, key.stop)

        key = timestamp_converter(key, self.tz)

        basemsg = "Getting {} but default attribute is not set".format(key)
        if self.empty:
            if self._has_default:
                return self.default
            else:
                if self.permissive:
                    return
                else:
                    raise KeyError(
                        "{} and timeseries is empty".format(basemsg))

        if key < self.lower_bound:
            if self._has_default:
                return self.default
            else:
                if self.permissive:
                    return
                else:
                    msg = "{}, can't deduce value before the oldest measurement"
                    raise KeyError(msg.format(basemsg))

        # If the key is already defined:
        if key in self.index:
            return self.data[key]

        if interpolate.lower() == "previous":
            fn = self._get_previous
        elif interpolate.lower() == "linear":
            fn = self._get_linear_interpolate
        else:
            raise ValueError("'{}' interpolation unknown.".format(interpolate))

        return fn(key)

    def _get_previous(self, time):
        # In this case, bisect_left == bisect_right == bisect
        # And idx > 0 as we already handled other cases
        previous_idx = self.data.bisect(time) - 1
        time_idx = self.index[previous_idx]
        return self.data[time_idx]

    def _get_linear_interpolate(self, time):
        # TODO: put it into a 'get_previous_index' method
        idx = self.data.bisect_left(time)
        previous_time_idx = self.index[idx - 1]

        # TODO: check on left bound case

        # out of right bound case:
        if idx == len(self):
            return self.data[previous_time_idx]

        next_time_idx = self.index[idx]

        previous_value = self.data[previous_time_idx]
        next_value = self.data[next_time_idx]

        coeff = (time - previous_time_idx) / (
            next_time_idx - previous_time_idx)

        value = previous_value + coeff * (next_value - previous_value)
        return value

    def slice(self, start, end):  # noqa A003
        """Slice your timeseries for give interval.

        Args:
            start (datetime or str): lower bound
            end (datetime or str): upper bound

        Returns:
            TimeSeries sliced
        """
        start = timestamp_converter(start, self.tz)
        end = timestamp_converter(end, self.tz)

        newts = TimeSeries(**self._kwargs_special_keys)

        for key in self.data.irange(start, end, inclusive=(True, False)):
            newts[key] = self[key]

        should_add_left_closure = (start not in newts.index
                                   and start >= self.lower_bound)
        if should_add_left_closure:
            newts[start] = self[start]  # is applying get_previous on self

        return newts

    def set_interval(self, start, end, value):
        """Set a value for an interval of time.

        Args:
            start (datetime or str): lower bound
            end (datetime or str): upper bound
            value: the value to be set

        Returns:
            self

        Raises:
            NotImplementedError: when no default is set.
        """
        if not self._has_default:
            msg = "At the moment, you have to set a default for set_interval"
            raise NotImplementedError(msg)

        start = timestamp_converter(start, self.tz)
        end = timestamp_converter(end, self.tz)

        keys = self.data.irange(start, end, inclusive=(True, False))

        last_value = self[end]

        for key in list(keys):
            del self.data[key]

        self[start] = value
        self[end] = last_value

    def compact(self):
        """Convert this instance to a compact version: consecutive measurement of the
        same value are discarded.

        Returns:
            TimeSeries
        """
        ts = TimeSeries(**self._kwargs_special_keys)
        for time, value in self.items():
            should_set_it = ts.empty or (ts[time] != value)
            if should_set_it:
                ts[time] = value
        return ts

    def iterintervals(self, end=None):
        """Iterator that contain start, end of intervals.

        Args:
            end (datetime): right bound of last interval.
        """
        lst_keys = SortedList(self.index)
        if not end:
            end = self.upper_bound
        else:
            end = timestamp_converter(end, self.tz)
            if end not in lst_keys:
                lst_keys.add(end)

        for i, key in enumerate(lst_keys[:-1]):
            next_key = lst_keys[i + 1]
            if next_key > end:  # stop there
                return
            yield key, next_key

    def equals(self, other, check_default=True, check_name=True):
        if not isinstance(other, self.__class__):
            raise TypeError("Can't compare {} with {}".format(
                self.__class__.__name__, other.__class__.__name__))

        is_equal = self.data == other.data

        if check_default:
            is_equal = is_equal and self.default == other.default

        if check_name:
            is_equal = is_equal and self.name == other.name

        return is_equal

    @property
    def tz(self):
        if self.empty:
            return pytz.UTC
        return str(self.index[0].tz)

    def tz_convert(self, tz):
        try:
            tz = pytz.timezone(tz)
        except pytz.UnknownTimeZoneError:
            raise ValueError('{} is not a valid timezone'.format(tz))

        ts = deepcopy(self)

        for key in ts.index:
            ts[key.tz_convert(tz)] = ts.data.pop(key)

        return ts



from .timeseries import TimeSeries


def assert_ts_equal(ts1, ts2, check_default=True, check_name=True):
    for ts in [ts1, ts2]:
        if not isinstance(ts, TimeSeries):
            msg = "{} is not of type TimeSeries"
            raise TypeError(msg.format(type(ts)))

    assert ts1.equals(ts2, check_default, check_name)



import operator

import pandas as pd


def timestamp_converter(ts, tz='UTC'):
    try:  # in case ts is a timestamp (also called epoch)
        ts = pd.to_datetime(float(ts), unit='ns')
    except Exception:
        ts = pd.Timestamp(ts)

    if not ts.tz:
        ts = ts.tz_localize(tz)
    return ts


MINTS = pd.Timestamp.min.tz_localize('UTC')
MAXTS = pd.Timestamp.max.tz_localize('UTC')


class NoDefault():
    def __repr__(self):
        return 'No default'


NO_DEFAULT = NoDefault()


def operation_factory(operation):
    def fn_operation(self, other):
        return self._operate(other, getattr(operator, operation))

    return fn_operation



import os

import pandas as pd
from pkg_resources import DistributionNotFound, get_distribution

from .timeseries import TimeSeries  # noqa: F401

try:
    _dist = get_distribution("ticts")

    # Normalize case for Windows systems
    dist_loc = os.path.normcase(_dist.location)
    here = os.path.normcase(__file__)

    if not here.startswith(os.path.join(dist_loc, "ticts")):
        # not installed, but there is another version that *is*
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = "Version not found."
else:
    __version__ = _dist.version


@pd.api.extensions.register_dataframe_accessor("to_ticts")
class FromPandasDTtoTicTS(object):
    def __init__(self, pandas_obj):
        self.obj = pandas_obj

    def __call__(self):
        return TimeSeries(self.obj)


@pd.api.extensions.register_series_accessor("to_ticts")
class FromPandasSeriestoTicTS(object):
    def __init__(self, pandas_obj):
        self.obj = pandas_obj

    def __call__(self):
        return TimeSeries(self.obj)



"""
isort: skip_file
"""

import logging

logger = logging.getLogger(__name__)

try:
    from bokeh.plotting import figure, show as bokeh_show
    BOKEH_IMPORTED = True
except ImportError:
    BOKEH_IMPORTED = False
    MSG_ERROR = ("'bokeh' is not installed. "
                 "Interactive Plot is not available. "
                 "Install it by using:\npip install bokeh")
    logger.debug(MSG_ERROR)


class TictsPlot:
    def _get_figure(self, title, dot_color, dot_size, **kwargs):
        kwargs = dict(
            title=title or self.name,
            x_axis_type="datetime",
            x_axis_label="timestamp",
            **kwargs)

        p = figure(**kwargs)

        p.step(
            self.index,
            self.values(),
            line_width=2,
            line_dash="4 4",
            mode="after")
        p.circle(
            self.index, self.values(), fill_color=dot_color, size=dot_size)

        return p

    def iplot(self,
              title=None,
              show=False,
              dot_color="red",
              dot_size=6,
              **kwargs):
        if not BOKEH_IMPORTED:
            raise ImportError(MSG_ERROR)

        fig = self._get_figure(
            title=title, dot_color=dot_color, dot_size=dot_size, **kwargs)

        if show:
            bokeh_show(fig)

        return fig
